/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 * @author Author: Sascha Nitsch (grumpydeveloper)
 **/

import { RenderTree } from './rendertree/rendertree';
import { RTPage, RTTransformation } from './rendertree/rtpage';

/**
 * configuration options for PresentationMaker
 */
export interface initialConfig {
  /** selector for source tree inside a &lt;script type="presentationmaker"&gt;...&lt;/script&gt; */
  sourceSelector?: string,
  /** alternative URL instead of script tag */
  sourceURI?: string,
  /** position to attach generated tree to */
  targetSelector: string,
  /** optional transition type, defaults to flip */
  transitionType?: string
  /** option ui selector string */
  uiSelector?: string;
}

/**
 * main class
 */
export class PresentationMaker {
  /** generated render tree */
  private tree: RenderTree;
  /** definitions for variable replacements */
  private definitions: Array<Map<string, string>>;
  /** custom modules defined by the import document */
  private modules: Map<string, RenderTree>;
  /** current page index */
  private currentPage: number;
  /** number of slide */
  private numPages: number;
  /** our viewport (window to the presentation) */
  private viewportElement: HTMLDivElement;
  /** transition type [flip / transition] */
  private transition: string;
  /** time for autohiding of UI */
  private timer: number;
  /** ui root node */
  private ui: HTMLDivElement;
  /** page selector */
  private select: HTMLSelectElement;
  /** X start position to detect slide left/right */
  private touchStartX: number;
  /** initial config from caller */
  private config: initialConfig;

  /**
   * change slide to given number
   * @param slideNum slide number 0-based
   */
  private changeSlide(slideNum: number) {
    if (slideNum < 0 && slideNum >= this.numPages) {
      return;
    }
    const newPage = this.tree.getElement(slideNum);
    const page = <RTPage>newPage;
    const pages = this.viewportElement.querySelectorAll('.page');
    if (this.currentPage >= 0 && this.currentPage < this.numPages) {
      (<HTMLElement>pages[this.currentPage]).classList.add('inactive');
    }
    switch (this.transition) {
      case 'transform':
        // this.transform = page.getView();
        this.currentPage = slideNum;
        this.moveViewpoint();
        break;
      default:
        if (this.currentPage >= 0 && this.currentPage < this.numPages) {
          (<HTMLElement>pages[this.currentPage]).style.display = 'none';
        }
        this.currentPage = slideNum;
        const newPage = <HTMLElement>pages[this.currentPage];
        newPage.style.display = 'block'
        break;
    }
    if (this.currentPage >= 0 && this.currentPage < this.numPages) {
      (<HTMLElement>pages[this.currentPage]).classList.remove('inactive');
    }
    document.location.hash = (this.currentPage + 1).toString();
    this.select.value = this.currentPage.toString();
    const prev = document.querySelector('[data-action="prev"]');
    if (this.currentPage == 0) {
      prev.classList.add('disabled');
    } else {
      prev.classList.remove('disabled');
    }
    const next = document.querySelector('[data-action="next"]');
    if (this.currentPage + 1 >= this.numPages) {
      next.classList.add('disabled');
    } else {
      next.classList.remove('disabled');
    }
  }

  /**
   * constructor
   */
  constructor(config: initialConfig) {
    if (config === undefined) {
      throw('please supply a configuration');
    }
    this.definitions = [];
    this.modules = new Map<string, RenderTree>();
    this.currentPage = 0;
    this.numPages = 0;
    this.transition = config.transitionType || 'flip';
    this.tree = new RenderTree(this, null);
    this.ui = document.querySelector(config.uiSelector || '.ui');
    this.select = this.ui.querySelector('select[data-id=pages]');
    if (document.location.hash) {
      this.currentPage = parseInt(document.location.hash.substring(1)) - 1;
    }
    this.config = config;
  }

  /**
   * create definitions from element tree
   * @param elem &lt;definition&gt;...&lt;/definition&gt; node tree
   */
  private createDefinitions(elem: HTMLElement): void {
    const definitions: Map<string, string> = new Map<string, string>();
    elem.childNodes.forEach((def) => {
      if (def.nodeType == 1) {
        const definitionElement = <HTMLElement>def;
        definitions.set(definitionElement.localName, definitionElement.innerHTML);
      }
    });
    this.definitions.push(definitions);
    
  }

  /**
   * create a custom module from input
   * @param elem &lt;module&gt;...&lt;/module&gt; node tree
   */
  private createModules(elem: HTMLElement): void {
    elem.childNodes.forEach((module) => {
      if (module.nodeType == 1) {
        const moduleElement = <HTMLElement>module;
        const tree = new RenderTree(this);
        tree.createFromElem(moduleElement);
        this.modules.set(moduleElement.localName, tree);
      }
    });
  }

  /**
   * create pages from input
   * @param elem &lt;slides&gt;...&lt;/slides&gt; node tree
   */
  private createPages(elem: HTMLElement): void {
    const pageElement = <HTMLElement>elem;
    this.tree.createFromElem(pageElement);
    this.numPages = this.tree.getElementCount();
    this.definitions[0].set('numpages', this.numPages.toString());
    this.definitions[0].set('page', '0');
    for (let i = 0; i < this.numPages; ++i) {
      const newPage = this.tree.getElement(i);
      const page = <RTPage>newPage;
      const option = document.createElement('option');
      option.value = i.toString();
      option.innerText = page.getName();
      this.select.append(option);
    }
  }

  /**
   * generate HTML tree from input XML
   */
  async generate(): Promise<void> {
    let root;
    const parser = new DOMParser();
    if (this.config.sourceSelector !== undefined) {
      const source = document.querySelector(this.config.sourceSelector);
      root = <HTMLElement>parser.parseFromString(source.outerHTML, 'application/xml').querySelector('script');
    } else {
      const response = await fetch(this.config.sourceURI);
      const data = await response.blob();
      root = <HTMLElement>parser.parseFromString(await data.text(), 'application/xml').querySelector('presentation');
    }

    // iterate over the elements
    root.childNodes.forEach((node) => {
      const elem = <HTMLElement>node;
      switch (elem.localName) {
        case 'definitions':
          this.createDefinitions(elem);
          break;
        case 'modules':
          this.createModules(elem);
          break;
        case 'pages':
          this.createPages(elem);
          break;
      }
      node = node.nextSibling;
    });
    // render pages
    this.viewportElement = <HTMLDivElement>document.createElement('div');
    this.viewportElement.id = 'viewport';
    this.tree.render(this.viewportElement);
    const renderTarget = <HTMLDivElement>document.querySelector(this.config.targetSelector);
    renderTarget.append(this.viewportElement);
  }

  /**
   * get a definition (value) for given name
   * @param name name to look up
   * @returns value or '' if not found
   * @internal
   */
  getDefinition(name: string): string {
    let ret = '';
    this.definitions.forEach((definitions) => {
      const i = definitions.get(name);
      if (i !== undefined) {
        ret = i;
      }
    });
    if (ret === '') {
      console.log('failed to get definition for', name);
    }
    return ret;
  }

  /**
   * create a new module (copy) with given name
   * @param name name of module
   * @return copy of module or null if not found
   * @internal
   */
  getNewModule(name: string): RenderTree {
    const i = this.modules.get(name);
    if (i !== undefined) {
      return (new RenderTree(this, i));
    }
    console.log('unknown render element', name);
    return null;
  }

  /**
   * handle mouse click or tap events
   * @param ev brower event
   */
  private handleClick = (ev?: MouseEvent) => {
    const target = <HTMLElement>ev.target;
    let action = target.getAttribute('data-action');
    if (action === null) {
      action = target.getAttribute("data-id");
    }
    let slideNum;
    switch (action) {
      case 'fullscreen':
        if (!document.fullscreenElement) {
          document.documentElement.requestFullscreen();
        } else {
          document.exitFullscreen();
        }
        ev.stopPropagation();
        ev.preventDefault();
        break;
      case 'prev':
        slideNum = Math.max(this.currentPage - 1, 0);
        break;
      case 'next':
      case null:
        slideNum = Math.min(this.currentPage + 1, this.numPages - 1);
        break;
    }
    if (slideNum !== undefined) {
      this.changeSlide(slideNum);
    }
    if (target.closest('.ui')) {
      ev.stopPropagation();
      ev.preventDefault();
    }
    return false;
  }

  /**
   * handle a change of the page dropdown
   */
  private handleDropdown = (ev?: Event) => {
    ev.stopPropagation();
    ev.preventDefault();
    this.changeSlide(parseInt(this.select.value));
    return false;
  }

  /**
   * handle key presses
   * @param ev pressed key
   */
  private handleKey = (ev?: KeyboardEvent) => {
    switch (ev.key) {
      case 'ArrowLeft':
        this.changeSlide(Math.max(this.currentPage - 1, 0));
        break;
      case 'ArrowRight':
        this.changeSlide(Math.min(this.currentPage + 1, this.numPages - 1));
        break;
    }
  }

  /**
   * handle mouse move and show ui
   */
  private handleMouseMove = () => {
    if (this.timer) {
      window.clearTimeout(this.timer);
    }
    this.ui.classList.add('visible');
    document.body.style.cursor = 'auto';
    this.timer = window.setTimeout(this.hideUI, 2000);
  }

  /**
   * handle end of touch event
   * creates slide left or right if swipe was long enough
   * @param ev browser touch event
   */
  private handleTouchEnd = (ev: TouchEvent) => {
    const touchEndX = ev.changedTouches[0].screenX;
    if (Math.abs(this.touchStartX - touchEndX) > 100) {
      let slideNum;
      if (this.touchStartX < touchEndX) {
        slideNum = Math.max(this.currentPage - 1, 0);
      } else {
        slideNum = Math.min(this.currentPage + 1, this.numPages - 1);
      }
      this.changeSlide(slideNum);
    } else {
      this.handleMouseMove();
    }
  }

  /**
   * handle start of touch
   * saves position to detect swipes
   */
  private handleTouchStart = (ev: TouchEvent) => {
    this.touchStartX = ev.changedTouches[0].screenX;
  }

  /**
   * hide ui
   * called by timer function
   */
  private hideUI = () => {
    this.ui.classList.remove('visible');
    document.body.style.cursor = 'none';
  }

  /**
   * move viewpoint to area defined by current page
   */
  private moveViewpoint = () => {
    const view = (<RTPage>this.tree.getElement(this.currentPage)).getView();
    let transform = '';
    if (view.viewRotation.x) {
      transform += ' rotateX(' + (-view.viewRotation.x) + 'deg) ' + transform;
    }
    if (view.viewRotation.y) {
      transform += ' rotateY(' + (-view.viewRotation.y) + 'deg) ' + transform;
    }
    if (view.viewRotation.z) {
      transform += ' rotateZ(' + (-view.viewRotation.z) + 'deg) ' + transform;
    }
    transform += ' translate3d(' + (-view.position.x) + 'px,' + (-view.position.y) + 'px, ' + (-view.position.z) + 'px)';

    // calculate zoom
    const wc = parseInt(this.getDefinition('width')) / view.size.w;
    const hc = parseInt(this.getDefinition('height')) / view.size.h;
    const scale = Math.min(wc, hc);
    const zoom = this.viewportElement.parentElement.clientHeight / parseFloat(this.getDefinition('height')) * scale;
    if (zoom != 1) {
      transform = 'scale(' + zoom + ')' + transform;
    }
    this.viewportElement.style.transform = transform;
  }

  /**
   * pop defintion list to remove last definition from stack
   * @internal
   */
  pop() {
    this.definitions.pop();
  }

  /**
   * add handles for presentation and start it
   */
  present() {
    window.addEventListener('resize', this.moveViewpoint);
    const list = document.querySelectorAll('body, [data-action]');
    list.forEach((e) => {
      e.addEventListener('click', this.handleClick);
    });
    window.addEventListener('keydown', this.handleKey);
    this.select.addEventListener('change', this.handleDropdown);
    window.addEventListener('mousemove', this.handleMouseMove);
    document.body.addEventListener('touchstart', this.handleTouchStart);
    document.body.addEventListener('touchend', this.handleTouchEnd);
    this.moveViewpoint();
    this.changeSlide(this.currentPage);
  }

  /**
   * push a new empty definition list to stack
   * @internal
   */
  push(): void {
    this.definitions.push(new Map());
  }

  /**
   * set a value to current definition list on stack
   * @internal
   */
  set(key: string, value: string): void {
    const definition = this.definitions.at(-1);
    definition.set(key, value);
  }
}
