/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 * @author Author: Sascha Nitsch (grumpydeveloper)
 * @internal
 **/

import { RTNode } from './rtnode';
import { RenderTree } from './rendertree';

/**
 * renders an image from <img> tag
 */
export class RTImg extends RTNode {
  /** src URI */
  private src: string;
  /** x position */
  private x: string;
  /** y position */
  private y: string;
  /** width */
  private width: string;
  /** height */
  private height: string;
  /** mask-image css string */
  private maskImage: string;

  /**
   * clone function
   * @return a copy of us
   */
  clone(): RTImg {
    return new RTImg(
      this.root,
      this.src,
      this.x,
      this.y,
      this.width,
      this.height,
      this.maskImage
    );
  }

  /**
   * constructor
   * @param root our parent RenderTree
   * @param src source URI
   * @param x x position
   * @param y y position
   * @param width image width
   * @param height image height
   * @param maskImage optional css for mask-image
   */
  constructor(root: RenderTree, src: string, x: string, y: string, width: string, height: string, maskImage: string) {
    super(root);
    this.src = src;
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
    this.maskImage = maskImage;
  }

  /**
   * create RTImg from given node into given RenderTree
   * @param root render tree to insert to
   * @param node input node
   * @returns new RTImg node
   */
  static fromElem(root: RenderTree, node: HTMLElement): RTImg {
    return new RTImg(
      root,
      this.getField(node, 'src'),
      this.getField(node, 'x'),
      this.getField(node, 'y'),
      this.getField(node, 'width'),
      this.getField(node, 'height'),
      this.getField(node, 'mask-image')
    );
  }

  /**
   * render image to HTML target
   * @param target target to render to
   */
  render(target: HTMLElement): void {
    const img = document.createElement('img');
    img.style.position = 'absolute';
    img.style.top = this.calc(this.resolveString(this.y)) + 'px';
    img.style.left = this.calc(this.resolveString(this.x)) + 'px';
    img.setAttribute('width', this.calc(this.resolveString(this.width)).toString());
    img.setAttribute('height', this.calc(this.resolveString(this.height)).toString());
    img.setAttribute('src', this.resolveString(this.src));
    if (this.maskImage) {
      img.style.maskImage = this.maskImage;
    }
    target.append(img);
  }
}
